@@ -4,14 +4,16 @@
 #pragma once
 
 #include "vec.h"
+#include "../traits/rktraits.h"
 
 namespace rkcommon {
   namespace math {
 
-    template <typename T>
+    template <typename T, typename = typename traits::is_arithmetic_t<T>>
     struct QuaternionT
     {
       using Vector = vec_t<T, 3>;
+      using Scalar = T;
 
       QuaternionT() {}
       QuaternionT(const QuaternionT &other)
@@ -69,6 +71,25 @@ namespace rkcommon {
       return QuaternionT<T>(a.r * b, a.i * b, a.j * b, a.k * b);
     }
 
+    template <typename T, typename U,
+        typename = traits::is_not_same_t<T, U>>
+    inline auto operator*(const T &a, const QuaternionT<U> &b)
+        -> QuaternionT<decltype(T() * U())>
+    {
+      using scalar_t = decltype(T() * U());
+      using quaternion_t = QuaternionT<scalar_t>;
+      return quaternion_t(scalar_t(a) * quaternion_t(b));
+    }
+
+    template <typename T, typename U, typename = traits::is_not_same_t<T, U>>
+    inline auto operator*(const QuaternionT<T> &a, const U &b)
+        -> QuaternionT<decltype(T() * U())>
+    {
+      using scalar_t = decltype(T() * U());
+      using quaternion_t = QuaternionT<scalar_t>;
+      return quaternion_t(quaternion_t(a) * scalar_t(b));
+    }
+
     template <typename T>
     inline QuaternionT<T> operator+(const QuaternionT<T> &a)
     {
@@ -95,7 +116,7 @@ namespace rkcommon {
       return conj(a) * rcp(a.r * a.r + a.i * a.i + a.j * a.j + a.k * a.k);
     }
     template <typename T>
-    inline float dot(const QuaternionT<T> &a, const QuaternionT<T> &b)
+    inline T dot(const QuaternionT<T> &a, const QuaternionT<T> &b)
     {
       return a.r * b.r + a.i * b.i + a.j * b.j + a.k * b.k;
     }
@@ -219,21 +240,21 @@ namespace rkcommon {
     inline typename QuaternionT<T>::Vector xfmPoint(
         const QuaternionT<T> &a, const typename QuaternionT<T>::Vector &b)
     {
-      return (a * QuaternionT<T>(b) * conj(a)).v();
+      return a * b;
     }
 
     template <typename T>
-    inline typename QuaternionT<T>::Vector xfmQuaternion(
-        const QuaternionT<T> &a, const typename QuaternionT<T>::Vector &b)
+    inline QuaternionT<T> xfmQuaternion(
+        const QuaternionT<T> &a, const QuaternionT<T> &b)
     {
-      return (a * QuaternionT<T>(b) * conj(a)).v();
+      return a * b;
     }
 
     template <typename T>
     inline typename QuaternionT<T>::Vector xfmNormal(
         const QuaternionT<T> &a, const typename QuaternionT<T>::Vector &b)
     {
-      return (a * QuaternionT<T>(b) * conj(a)).v();
+      return a * b;
     }
 
     template <typename T>
@@ -248,21 +269,21 @@ namespace rkcommon {
       return a.r != b.r || a.i != b.i || a.j != b.j || a.k != b.k;
     }
 
-    template <typename T>
-    QuaternionT<T>::QuaternionT(const typename QuaternionT<T>::Vector &vx,
-                                const typename QuaternionT<T>::Vector &vy,
-                                const typename QuaternionT<T>::Vector &vz)
+    template <typename T, typename U>
+    QuaternionT<T, U>::QuaternionT(const typename QuaternionT<T, U>::Vector &vx,
+                                   const typename QuaternionT<T, U>::Vector &vy,
+                                   const typename QuaternionT<T, U>::Vector &vz)
     {
       if (vx.x + vy.y + vz.z >= T(zero)) {
         const T t = T(one) + (vx.x + vy.y + vz.z);
-        const T s = rsqrt(t) * T(0.5f);
+        const T s = rsqrt(t) * T(.5);
         r         = t * s;
         i         = (vy.z - vz.y) * s;
         j         = (vz.x - vx.z) * s;
         k         = (vx.y - vy.x) * s;
       } else if (vx.x >= max(vy.y, vz.z)) {
         const T t = (T(one) + vx.x) - (vy.y + vz.z);
-        const T s = rsqrt(t) * T(0.5f);
+        const T s = rsqrt(t) * T(.5);
         r         = (vy.z - vz.y) * s;
         i         = t * s;
         j         = (vx.y + vy.x) * s;
@@ -270,7 +291,7 @@ namespace rkcommon {
       } else if (vy.y >= vz.z)  // if ( vy.y >= max(vz.z, vx.x) )
       {
         const T t = (T(one) + vy.y) - (vz.z + vx.x);
-        const T s = rsqrt(t) * T(0.5f);
+        const T s = rsqrt(t) * T(.5);
         r         = (vz.x - vx.z) * s;
         i         = (vx.y + vy.x) * s;
         j         = t * s;
@@ -278,7 +299,7 @@ namespace rkcommon {
       } else  // if ( vz.z >= max(vy.y, vx.x) )
       {
         const T t = (T(one) + vz.z) - (vx.x + vy.y);
-        const T s = rsqrt(t) * T(0.5f);
+        const T s = rsqrt(t) * T(.5);
         r         = (vx.y - vy.x) * s;
         i         = (vz.x + vx.z) * s;
         j         = (vy.z + vz.y) * s;
@@ -286,15 +307,15 @@ namespace rkcommon {
       }
     }
 
-    template <typename T>
-    QuaternionT<T>::QuaternionT(const T &yaw, const T &pitch, const T &roll)
+    template <typename T, typename U>
+    QuaternionT<T, U>::QuaternionT(const T &yaw, const T &pitch, const T &roll)
     {
-      const T cya = cos(yaw * T(0.5f));
-      const T cpi = cos(pitch * T(0.5f));
-      const T cro = cos(roll * T(0.5f));
-      const T sya = sin(yaw * T(0.5f));
-      const T spi = sin(pitch * T(0.5f));
-      const T sro = sin(roll * T(0.5f));
+      const T cya = cos(yaw * T(.5));
+      const T cpi = cos(pitch * T(.5));
+      const T cro = cos(roll * T(.5));
+      const T sya = sin(yaw * T(.5));
+      const T spi = sin(pitch * T(.5));
+      const T sro = sin(roll * T(.5));
       r           = cro * cya * cpi + sro * sya * spi;
       i           = cro * cya * spi + sro * sya * cpi;
       j           = cro * sya * cpi - sro * cya * spi;
@@ -313,8 +334,8 @@ namespace rkcommon {
     inline QuaternionT<T> slerp(const float factor, const QuaternionT<T> &_a, const QuaternionT<T> &b)
     {
       QuaternionT<T> a(_a);
-      float d = dot(a, b);
-      if (d < 0.f) { // prevent "long way around"
+      T d = dot(a, b);
+      if (d < 0.) { // prevent "long way around"
         a = -a;
         d = -d;
       }
@@ -322,10 +343,10 @@ namespace rkcommon {
         return normalize(rkcommon::math::lerp(factor, a, b));
       }
 
-      const float theta0 = std::acos(d);
-      const float theta = theta0 * factor;
-      const float fb = std::sin(theta) / std::sin(theta0);
-      const float fa = std::cos(theta) - d * fb;
+      const T theta0 = std::acos(d);
+      const T theta = theta0 * factor;
+      const T fb = std::sin(theta) / std::sin(theta0);
+      const T fa = std::cos(theta) - d * fb;
 
       return fa * a + fb * b;
     }
